
/*
 * vserver stress tests
 *
 * Copyright (C) 2004 Herbert Poetzl <herbert@13thfloor.at>
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * V0.1 basic SMP test framework
 * V0.2 per thread data
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sched.h>
#include <string.h>

#include <sys/time.h>
#include <sys/poll.h>

#include <asm/atomic.h>

#include "affinity.h"
#include "timing.h"
#include "tests.h"


#define VERSION "V0.2"


int 	verb_lvl = 1;
struct	timeval main_start;


static	char	opt_notcpu0 = 0;
static	char	opt_affinity = 0;

static	char	opt_test[10];
static	char	*test_arg[10];

static	int 	num_cpu = 1;
static	int 	num_cycles = 1;
static	int 	num_thread = 0;
static	int 	num_offset = 0;
static	int 	num_cpuoff = 0;

static	long 	adjust = 0;

static  char 	*cmd_name = NULL;

static	volatile
enum 	_sync_cmd {
    	CMD_SPIN = 0,
	CMD_RUN,
	CMD_ABORT,
}   	cmd;

static  atomic_t active;
static  atomic_t waiting;

static	per_thread _main_ptd = {
    	.index = -1,	
    	};

static	per_thread *ptd = &_main_ptd;


#define PTD_SIZE  sizeof(per_thread)

#define CLONE_FLAGS 	( CLONE_PARENT | CLONE_FS | CLONE_VM  \
    	    	    	| CLONE_THREAD | CLONE_SIGHAND )


long gnow(per_thread *ptd)
{
    	long delta;

    	if (!ptd)
	    ptd = &_main_ptd;    	
	tm_start_time(&ptd->now);
	delta = tm_delta_ms(&main_start, &ptd->now);
	return delta;
}

long rnow(per_thread *ptd, struct timeval *ref)
{
    	long delta;
    	
	tm_start_time(&ptd->now);
	delta = tm_delta_ms(ref, &ptd->now);
	return delta;
}

#define CLONE_PARENT	0x00008000	/* same parent as the cloner */
#define CLONE_THREAD	0x00010000	/* Same thread group? */



void 	do_nothing(int signum)  { }

void 	do_abort(int signum) 	{ cmd = CMD_ABORT; }


#define DO_TEST(r, n, test) do {    	    	\
 	if (opt_test[n])    	    	    	\
	    r = test(ptd, test_arg[n]);      	\
	} while (0)
   	

int 	do_test_cycle(per_thread *ptd) 
{
    	int ret = 0;
	int index = ptd->index;
	int cycle = ptd->cycle;
    	long delta;
    	
	msleep(index*num_offset);
    	XPRINTF(0,_G,_T, "starting test cycle %d ...\n", cycle);
	tm_start_time(&ptd->tv_cycle);
	
    	DO_TEST(ret, 0, test_cpuhog);
    	DO_TEST(ret, 1, test_syshog);

    	delta = tm_delta_time(&ptd->tv_cycle);
    	XPRINTF(0,_G,_T, "finished test cycle %d.\n", cycle);
	XPRINTF(0,_G,_T, "cycle time %ld.%03ldms\n", DSPLIT(delta));
	return 0;
}

int 	do_cpu_thread(void *arg) 
{
	per_thread *ptd = arg;
	int count = 0;
	int cpu = ptd->cpu;
    	unsigned long mask = 1 << cpu;
    	unsigned int len = sizeof(mask);
	
	tm_start_time(&ptd->tv_start);

	if (opt_affinity) {
	    TPRINTF(0, "changing cpu affinity ...\n");
	    if (sched_setaffinity(0, len, &mask) < 0)
 	    	TPRINTF(0, "setaffinity: %s.\n", strerror(errno));
	    if (sched_getaffinity(0, len, &mask) < 0)
 	    	TPRINTF(0, "getaffinity: %s.\n", strerror(errno));
	    if (mask ^ (1 << cpu)) {
	    	TPRINTF(0, "wrong affinity, exiting.\n");
	    	exit(1);
	    }
    	}

    	while (count < num_cycles) {
	    ptd->cycle = count;
	    
    	    TPRINTF(0, "ready and waiting ...\n");
    	    atomic_inc(&waiting);
    	    while (cmd == CMD_SPIN);
	    msleep(cpu*num_cpuoff);
    	    atomic_inc(&active);
    	    do_test_cycle(ptd);
    	    atomic_dec(&active);
	    atomic_dec(&waiting);
	    
	    while (cmd == CMD_RUN)
	    	msleep(100);
	    if (cmd == CMD_ABORT)
	    	break;
    	    count++;
	}
	return 0;
}


#define OPTIONS "hauC:U:N:O:T:0:1:2:3:4:5:6:7:8:9:"

int	main(int argc, char *argv[])
{
        extern int optind;
        extern char *optarg;

	struct sigaction action;
        char c, errflg = 0;
	int ret, i, j, threads;

        cmd_name = argv[0];
        while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
            switch (c) {
            case 'h':
                fprintf(stderr,
                    "This is %s " VERSION "\n"
                    "options are:\n"
                    "-h        print this help message\n"
                    "-# <arg>  activate test # with <arg>\n"
		    "-a        set cpu affinity for threads\n"
		    "-u        do not use first cpu\n"
                    "-C <num>  number of test cycles\n"
 		    "-U <num>  number of CPUs to test\n"
 		    "-N <num>  number of threads per cpu\n"
		    "-O <num>  offset CPUs by <num> msecs\n"
		    "-T <num>  offset threads by <num> msecs\n"
                    ,cmd_name);
                exit(0);
                break;
		
	    case 'a':
		opt_affinity = 1;
	        break;
	    case 'u':
		opt_notcpu0 = 1;
	        break;
            case 'C':
                num_cycles = atol(optarg);
                break;
	    case 'U':
		num_cpu = atol(optarg);
	        break;
	    case 'N':
		num_thread = atol(optarg);
	        break;
            case 'O':
                num_cpuoff = atol(optarg);
                break;
            case 'T':
                num_offset = atol(optarg);
                break;
		
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	    	i = (c - '0');
		opt_test[i] = 1;
		test_arg[i] = optarg;
		break;
		
            case '?':
            default:
                errflg++;
                break;
            }
        }
        if (errflg) {
            fprintf(stderr, 
                "Usage: %s -" OPTIONS "\n"
                "%s -h for help.\n",
                cmd_name, cmd_name);
            exit(2);
        }

	adjust = tm_calc_overhead(100);

    	/* block SIGALRM */
	action.sa_handler = do_nothing;
	action.sa_flags = 0;
	sigaction(SIGALRM, &action, NULL);

    	/* block SIGINT */
	action.sa_handler = do_abort;
	action.sa_flags = 0;
	sigaction(SIGINT, &action, NULL);

	tm_start_time(&main_start);
    	atomic_set(&active, 0);
    	atomic_set(&waiting, 0);
    	cmd = CMD_SPIN;
    	threads = 0;
	
	for (i=(opt_notcpu0)?1:0; i<num_cpu; i++) {
	    void *stack;
	    per_thread *pt;
	
	    for (j=0; j<num_thread; j++) {
		DPRINTF(0, "spawning thread %d%d\n", i, j);
		stack = calloc(STACK_SIZE + PTD_SIZE, 1);
		if (stack) {
		    stack += STACK_SIZE;
		    pt = stack;
		    pt->index = j;
		    pt->cpu = i;
		    ret = clone(do_cpu_thread, stack,
			CLONE_FLAGS, (void *)pt);
		    if (ret < 0)
			DPRINTF(0, "clone: %s\n", strerror(errno));
		    else
		    	threads++;
		} else
		    DPRINTF(0, "no stack for %d :/\n", i);
    	    }
    	}

	for (i=0; i<num_cycles; i++) {
    	    while (atomic_read(&waiting) < threads)
	    	msleep(100);
	    cmd = CMD_RUN;
	    while (atomic_read(&waiting))
	    	msleep(100);
    	    DPRINTF(0, "finished test cycle %d.\n", i);
    	    if (cmd == CMD_ABORT)
	    	break;
	    cmd = CMD_SPIN;
	}
		
	if (cmd == CMD_ABORT) {
	    DPRINTF(0, "test run aborted by user.\n");
	    errflg = 1;
	}
	exit((errflg)?1:0);
}

